/* $Workfile: DspDemo2.c $       $Revision: 13 $      $Modtime: 1/24/01 5:58p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    DspDemo2.c

    A simple example console application demonstrating the use of the DSP SDK
    example program DspDemo2.out. This program works with the OTX PCI adapters and
    ASM Daughter Boards, such as:
      THOR-2-PCI + Vidar-5x4-ASM
      Arni-32-PCI + Vidar-5x4-ASM
      Balder-8S-PCI + Vidar-5x4-ASM
      Vidar-5x16-PCI.

    The example program opens 4 devices and loads the user written DSP SDK
    Example program "DspDemo2.out" to these 4 DSP devices.

    A codec (handset) is connected to each DSP. The DSP code loops the handset
    microphone to the handset speaker. The demo program utilizes the I/O
    Control interface to control speaker gain. Simple keystroke commands are
    used to vary the speaker gain.

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <time.h>

#ifdef WIN32 // Testing sound
#include <windows.h>
#include <mmsystem.h>
#endif

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxErr.h"
#include "OtxBrd.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxDsp.h"



#include <Demo2Io.h>   // IO Control codes for the DSP to host communication
#include <DspIoctl.h>  // IO Control codes for the DSP to host communication

#define DEMO_TASK_REF_DEBUG 501   // Task Reference for a non-blocking call
#define DEMO_DSP_MAX 4

#define DEMO_DEFAULT_LOG_FILE "DspDemo2.log"
#define DEMO_DEFAULT_BOARD_NO 0

typedef struct DemoEnvT {
  char           *logFileName;
  enum OtxPhysicalDeviceTypeE nBoardType;
  OTX_INT32      nDsps;
  OTX_BOOL       bAsmPresent;
  int            boardNo;
} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoOpenPhysicalDevices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoCreateLogicalDevices(void);
static void DemoEnableLogicalDevices(void);
static void DemoConnectDevices(void);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);
static void DemoPrintTimeStamp(OTX_DATETIME nTimeStamp);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static void DemoLogf(char *formatStr, ...);
static void DemoHelpPanel(void);
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);


/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;


struct {
  OTX_HANDLE hBoardController0;
  OTX_HANDLE hTss;
  OTX_HANDLE hAsmController0;
  OTX_HANDLE hDsp[DEMO_DSP_MAX];
  OTX_HANDLE hCodec[4];
} g_hPhysDevs;

struct {
  OTX_HANDLE hUserApp[DEMO_DSP_MAX];
} g_hLogicDevs;

struct {
  OTX_HANDLE hDspCodec[4];
  OTX_HANDLE hCodecDsp[4];
} g_hPipes;



/*******************************************************************************
*******************************************************************************/


void DemoCreateEventQueue(void)
/*******************************************************************************
  @func
    Create the event queue to be used by the driver to inform the application
    about events, such as line status changes, or detected DMTF digits.
*******************************************************************************/
{
  OTX_RESULT result;

  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent);
  if (result == OTX_E_NOTIMPL) {
    g_hNotificationEvent = NULL;    // D16V86 apps cannot use event objects
    result = OTX_S_OK;
  }

  DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(g_hEventQueue, result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC,0); DemoCheckRc(g_hEventQueue, result);
}



void DemoOpenPhysicalDevices(void)
/*******************************************************************************
  @func
    Open the Physical Devices on the Thor-2-PCI board
  @comm
    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32  i;

  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
    case OTX_DEVICE_ARNI_PCI:
      DemoLogf("Opening OTX_DEVICE_ARNI_PCI board\n");
      break;
    case OTX_DEVICE_THOR_PCI:
      DemoLogf("Opening OTX_DEVICE_THOR_PCI board\n");
      break;
    case OTX_DEVICE_VIDAR_5x_PCI:
      DemoLogf("Opening OTX_DEVICE_VIDAR_5x_PCI board\n");
      break;
    case OTX_DEVICE_BALDER_S_PCI:
      DemoLogf("Opening OTX_DEVICE_BALDER_S_PCI board\n");
      break;
    default:
      DemoLogf("Opening '%d' board\n", g_demoEnv.nBoardType);
      break;
  }
  result = OtxDrvOpenPhysicalDevice(0, g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hBoardController0); DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  // Get a handle to the Time-Space Switch (TSS) on the board
  DemoLogf("Open OTX_DEVICE_TSS\n");
  result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController0, OTX_DEVICE_TSS, 0, 0, g_hEventQueue, &g_hPhysDevs.hTss); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
  DemoLogf("Enable OTX_DEVICE_TSS\n");
  result = OtxDrvEnable(g_hPhysDevs.hTss,OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hTss, result);

  // Get handles to Codec devices #0 - #3
  for (i = 0; i < 4; i++) {
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController0, OTX_DEVICE_CODEC, i, i, g_hEventQueue, &g_hPhysDevs.hCodec[i]); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
  }


  if ( g_demoEnv.nBoardType == OTX_DEVICE_VIDAR_5x_PCI) {
    // On Vidar-5x16-PCI, DSPs are owned by the NIC Board Device
    for (i = 0; i < g_demoEnv.nDsps; i++) {
      DemoLogf("Opening DSP#%d\n", i);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController0, OTX_DEVICE_DSP, i, i, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    }

  } else {
    // Get a handle to to the ASM Daughter Board
    DemoLogf("Opening OTX_DEVICE_VIDAR_5x_ASM board\n");
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController0, OTX_DEVICE_VIDAR_5x_ASM, 0, 0, g_hEventQueue, &g_hPhysDevs.hAsmController0); DemoCheckRc(g_hPhysDevs.hBoardController0, result);

    for (i = 0; i < g_demoEnv.nDsps; i++) {
      DemoLogf("Opening DSP#%d\n", i);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hAsmController0, OTX_DEVICE_DSP, i, i, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hAsmController0, result);
    }

  }

  // Load DSP Program
  for (i = 0; i < g_demoEnv.nDsps; i++) {
    DemoLogf("Loading program to DSP#%d\n", i);
    result = OtxDspRunProgram(g_hPhysDevs.hDsp[i], "DspDemo2.out", "_c_int00"); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
  }

}



void DemoEnablePhysicalDevices(void)
/*******************************************************************************
  @func
    Enable Configured Physical Devices
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32  i;

  result = OtxDrvEnable(g_hPhysDevs.hBoardController0,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
  if (g_demoEnv.bAsmPresent) {
    result = OtxDrvEnable(g_hPhysDevs.hAsmController0, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hAsmController0, result);
  }

  for (i = 0; i < g_demoEnv.nDsps; i++) {
    result = OtxDrvEnable(g_hPhysDevs.hDsp[i],            OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
  }
}



void DemoCreateLogicalDevices(void)
/*******************************************************************************
  @func
    Create the needed Logical Devices
  @comm
    The logical devices are differenciated from the phyiscal devices by the fact
    the number of logical devices is not set. More of them can be requested
    by the user and allocated by the system if needed.
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32  i;

  for (i = 0; i < g_demoEnv.nDsps; i++) {
    // Create one User Application Logical Devices per DSP
    result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hDsp[i], OTX_LDEVICE_USER_APPLICATION, g_hEventQueue, i, &g_hLogicDevs.hUserApp[i]); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
  }
}



void DemoEnableLogicalDevices(void)
/*******************************************************************************
  @func
    Enable Configured Logical Devices
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32  i;

  for (i = 0; i < g_demoEnv.nDsps; i++) {
    result = OtxDrvEnable(g_hLogicDevs.hUserApp[i],   OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hUserApp[i], result);
  }
}


void DemoConnectDevices(void)
/*******************************************************************************
  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.
  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset of a
    Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a characterized
    bu its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32 i = 0;

  for (i = 0; i < g_demoEnv.nDsps; i++) {
    // Create voice pipes for DspX <--> CodecX
    result = OtxDrvCreatePipe(64, &g_hPipes.hDspCodec[i]); DemoCheckRc(g_hPipes.hDspCodec[i], result);
    result = OtxDrvCreatePipe(64, &g_hPipes.hCodecDsp[i]); DemoCheckRc(g_hPipes.hCodecDsp[i], result);

#if 0 // Letting the OtxDriver pick the timeslot for the DSP highway
    // Connect Codec Microphone to DSP.
    result = OtxDrvConnectPipe(g_hPipes.hCodecDsp[i], OTX_PIPE_INPUT,  g_hPhysDevs.hCodec[i], -1, NULL, 0); DemoCheckRc(g_hPipes.hCodecDsp[i], result);
    result = OtxDrvConnectPipe(g_hPipes.hCodecDsp[i], OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[i], -1, NULL, 0); DemoCheckRc(g_hPipes.hCodecDsp[i], result);
    result = OtxDrvConnectLogicalDevice(g_hPipes.hCodecDsp[i], -1, g_hLogicDevs.hUserApp[i],   -1); DemoCheckRc(g_hPipes.hCodecDsp[i], result);

    // Connect DSP.to Codec Speaker
    result = OtxDrvConnectPipe(g_hPipes.hDspCodec[i], OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[i], -1, NULL, 0); DemoCheckRc(g_hPipes.hDspCodec[i], result);
    result = OtxDrvConnectPipe(g_hPipes.hDspCodec[i], OTX_PIPE_OUTPUT, g_hPhysDevs.hCodec[i], -1, NULL, 0); DemoCheckRc(g_hPipes.hDspCodec[i], result);
    result = OtxDrvConnectLogicalDevice(g_hLogicDevs.hUserApp[i], -1, g_hPipes.hDspCodec[i], -1); DemoCheckRc(g_hLogicDevs.hUserApp[i], result);
#else // Specific timeslots on the DSP highway
    {
      OTX_UINT8 sqTimeSlotMask[32];
      OTX_INT32 ts = 1;

      memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      sqTimeSlotMask[ts] = 0xFF;

      // Connect Codec Microphone to DSP.
      result = OtxDrvConnectPipe(g_hPipes.hCodecDsp[i], OTX_PIPE_INPUT,  g_hPhysDevs.hCodec[i], -1, NULL, 0); DemoCheckRc(g_hPipes.hCodecDsp[i], result);
      result = OtxDrvConnectPipe(g_hPipes.hCodecDsp[i], OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[i], -1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hCodecDsp[i], result);
      result = OtxDrvConnectLogicalDevice(g_hPipes.hCodecDsp[i], -1, g_hLogicDevs.hUserApp[i],   -1); DemoCheckRc(g_hPipes.hCodecDsp[i], result);

      // Connect DSP.to Codec Speaker
      result = OtxDrvConnectPipe(g_hPipes.hDspCodec[i], OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[i], -1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hDspCodec[i], result);
      result = OtxDrvConnectPipe(g_hPipes.hDspCodec[i], OTX_PIPE_OUTPUT, g_hPhysDevs.hCodec[i], -1, NULL, 0); DemoCheckRc(g_hPipes.hDspCodec[i], result);
      result = OtxDrvConnectLogicalDevice(g_hLogicDevs.hUserApp[i], -1, g_hPipes.hDspCodec[i], -1); DemoCheckRc(g_hLogicDevs.hUserApp[i], result);
    }
#endif
  }

}




static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  short c;
  if ( OTX_FAILURE(result) ) {

    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX) \n", result);
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      eDeviceType = 999;    // Device type for hDevice 
      OTX_UINT32      nDeviceNumber = 999;  // Device number for hDevice
      
      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType); 
      } 
      
      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber); 
      }
      
      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(hDevice, result, 256, &msgStr[0]);
      if (OTX_SUCCESS(otxRc)) {
        DemoLogf("ERROR(%#lX) Dev(%X,%d,%d): %s \n", result, hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
      } else {
        DemoLogf("ERROR(%#lX) Dev(%X,%d,%d) Cannot translate this result code \n", result, hDevice, (int) eDeviceType, (int)nDeviceNumber);
      }
    }


    DemoLogf("Do you want to exit (Y/N) ");
    c = getch();
    if (c == 'y' || c == 'Y') {
      // Clean up
      OtxDrvReleaseLib();
      OsaConsoleExit();
      exit(1);
    }

  }
}



/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct tm *ptCurrentDayAndTime;
  time_t nCurrentTimeSecs;
  
  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 1164447360 * 100000000;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
  #endif
  
  // Convert to seconds since Jan 1, 1970 (so we can use this time with localtime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
  
  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);
  
  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = localtime(&nCurrentTimeSecs);
  
  if (ptCurrentDayAndTime) {
    
    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ", 
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min, 
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }
  
  
}



void *DemoIdleFunction(
/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
  void *lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;


  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
    //DemoLogf("Event signalled\n");
    do {
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(g_hEventQueue, result);

      if (result == OTX_S_OK) {
        // We have a driver event
        // Printf the event data.
        // In a real application branch based on the data
        DemoPrintEvent(&eventData);

      }
    } while (result == OTX_S_OK);
  }

  return 0;


}



void DemoPrintEvent(
/*******************************************************************************
  @func
    Prints a Driver Event.
*******************************************************************************/
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  char msgStr[256];


  switch (pEventData->m_eDeviceType) {
  case OTX_DEVICE_TSS:
    // Event from the Time-Space Switch
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("TSS Event: %s \n", msgStr);
    } else {
      DemoLogf("TSS Invalid Event: %ld \n", pEventData->m_nCode);
    }
    break;

  case OTX_DEVICE_DSP:
    // Non-logical device related event from the DSPs
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("DSP[%d] Event: %s (nRequestId=%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nRequestId);
    } else {
      DemoLogf("DSP[%d] Invalid Event: %ld (nRequestId=%d)\n", pEventData->m_nSourceId, pEventData->m_nCode, pEventData->m_nRequestId);
    }
    break;

  case OTX_LDEVICE_USER_APPLICATION:
    // Events from the Demo2.out logical device
    if (pEventData->m_nCode == OTX_USER_APPLICATION_EC_READ_DATA) {
      OTX_UINT8  buf[8196];
      OTX_UINT32 nBytesReturned;
      OTX_UINT32 nIoControlCode;

      otxRc = OtxDspReadData(pEventData->m_hDevice, OTX_TASK_SYNC, buf, 8196, &nBytesReturned, &nIoControlCode);
      if (otxRc == OTX_S_OK) {
        switch(nIoControlCode) {
        case DEMO2_NOTIFY_DEBUG_DATA:
          {
            Demo2NotifyGetDebugDataS *pDataS = (Demo2NotifyGetDebugDataS *)buf;

            DemoPrintTimeStamp(pEventData->m_nTimeStamp);
            if (nBytesReturned != sizeof(Demo2NotifyGetDebugDataS)) {
              DemoLogf("Unexpected data length for this IoControlCode (nIoControlCode = %d, nBytesReturned=%d)\n", nIoControlCode, nBytesReturned);
            } else {
              DemoLogf("DEMO2_NOTIFY_DEBUG_DATA[%d] (nRequestId=%d, nBytes=%d)\n", pEventData->m_nSourceId, pEventData->m_nRequestId, nBytesReturned);
              DemoLogf("nData0 = %.4X\n", pDataS->nData0);
              DemoLogf("nData1 = %.4X\n", pDataS->nData1);
              DemoLogf("nData2 = %.4X\n", pDataS->nData2);
            }
          }
          break;
        case DEMO2_NOTIFY_STREAM_RX_DATA:
          {
            //Demo2NotifyStreamRxDataS *pDataS = (Demo2NotifyStreamRxDataS *)buf;
            if (nBytesReturned != sizeof(Demo2NotifyStreamRxDataS)) {
              DemoLogf("Unexpected data length for this IoControlCode (nIoControlCode = %d, nBytesReturned=%d)\n", nIoControlCode, nBytesReturned);
            } else {
              static OTX_INT32 i = 0;

              if ((i++ % 500) == 0) {
                DemoLogf("RX DATA[%d] (nRequestId=%d, nBytes=%d)\n", pEventData->m_nSourceId, pEventData->m_nRequestId, nBytesReturned);
              }
#if 0
              for (i = 0; i < 25; i++) {
                DemoLogf("%.4X ", pDataS->nData[i]);
              }
              DemoLogf("\n");
#endif
            }
          }
          break;
        default:
          DemoLogf("Unknown IoControlCode received in OtsDspReadData (nIoControlCode = %d, nBytesReturned=%d, nRequestId=%d)\n", nIoControlCode, nBytesReturned, pEventData->m_nRequestId);
          break;
        }
      } else {
        OTX_RESULT errorCode = otxRc;
        DemoPrintTimeStamp(pEventData->m_nTimeStamp);
        otxRc = OtxDrvResultCode2String(pEventData->m_hDevice, errorCode, 256, &msgStr[0]);
        if (OTX_SUCCESS(otxRc)) {
          DemoLogf("User App[%d] Error: %s \n", pEventData->m_nSourceId, msgStr);
        } else {
          DemoLogf("User App[%d] Invalid Error Code: %ld\n", pEventData->m_nSourceId, errorCode);
        }
      }

    } else {

      DemoPrintTimeStamp(pEventData->m_nTimeStamp);
      otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
      if (OTX_SUCCESS(otxRc)) {
        DemoLogf("User App[%d] Event: %s (nRequestId=%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nRequestId);
      } else {
        DemoLogf("User App[%d] Invalid Event: %ld (nRequestId=%d)\n", pEventData->m_nSourceId, pEventData->m_nCode, pEventData->m_nRequestId);
      }
    }
    break;

  default:
    // Unhandled event
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);
    DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }


}


OTX_INT32 DemoProcessUserCommand(char commandChar)
/*******************************************************************************
  @func
    Processes User Commands
*******************************************************************************/
{
  static OTX_INT32      nCurrentUserApp = 0;
  static OTX_BOOL       bStreamRxData[DEMO_DSP_MAX];
  OTX_RESULT result;

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    DemoLogf(" g   Set Handset Speaker Gain for current DSP User Application \n");
    DemoLogf(" e   Enable Handset Mic->Speaker Loop on current Dsp\n");
    DemoLogf(" E   Disable Handset Mic->Speaker Loop on current Dsp\n");
    DemoLogf(" d   Get Debug data from the DSP User Application\n");
    DemoLogf(" s   Get Expanded sample value (16-bits) from the Handset Mic from the DSP User Application\n");
    DemoLogf(" t   Enable/Disable (toggle) continous stream of the received data stream from the Handset Mic\n");
    DemoLogf(" 0-3 Select active DSP User Application, which is then used for subsequent commands\n");
    break;

/*
  case 'a':
    {
      OTX_INT32 i;
      OTX_INT32 j;

      DemoLogf("Allocate 10M of buffers\n");
      for (j = 0; j < 10; j++) {
        void *pDummy;
        for (i = 0; i < 1024; i++) {
          pDummy = malloc(1024);
          if (!pDummy) {
            DemoLogf("Out of memory\n");
            j=10;
            break;
          }
        }
      }
      DemoLogf("Done\n");
    }
    break;
*/

  case 'm':
    {
      OTX_HANDLE hHs3MicPipe;
      OTX_HANDLE hHs2MicPipe;

      // Create voice pipes for Hs2 and Hs3
      result = OtxDrvCreatePipe(64, &hHs3MicPipe); DemoCheckRc(hHs3MicPipe, result);
      result = OtxDrvCreatePipe(64, &hHs2MicPipe); DemoCheckRc(hHs2MicPipe, result);

      // Connect Hs2 <--> Hs3 in the time-space switch
      result = OtxDrvConnectPipe(hHs2MicPipe, OTX_PIPE_INPUT,  g_hPhysDevs.hCodec[2], -1, NULL, 0); DemoCheckRc(hHs2MicPipe, result);
      result = OtxDrvConnectPipe(hHs2MicPipe, OTX_PIPE_OUTPUT, g_hPhysDevs.hCodec[3], -1, NULL, 0); DemoCheckRc(hHs2MicPipe, result);

      result = OtxDrvConnectPipe(hHs3MicPipe, OTX_PIPE_INPUT,  g_hPhysDevs.hCodec[3], -1, NULL, 0); DemoCheckRc(hHs3MicPipe, result);
      result = OtxDrvConnectPipe(hHs3MicPipe, OTX_PIPE_OUTPUT, g_hPhysDevs.hCodec[2], -1, NULL, 0); DemoCheckRc(hHs3MicPipe, result);
    }
    break;

  case 't':
    {
      Demo2IoInStreamRxDataS sIn;
      OTX_UINT32 nBytesReturned;
      OTX_UINT16 nDemo2Rc;
      
      bStreamRxData[nCurrentUserApp] = !bStreamRxData[nCurrentUserApp];
      
      DemoLogf("%s stream of RX data for User App #%d ...", bStreamRxData[nCurrentUserApp] ? "Enabling" : "Disabling", nCurrentUserApp);
      
      sIn.bStreamRxData = (OTX_INT16) bStreamRxData;
      
      result = OtxDspIoControl(g_hLogicDevs.hUserApp[nCurrentUserApp], DEMO2_IOCTL_STREAM_RX_DATA, &sIn, sizeof(sIn), &nDemo2Rc, sizeof(nDemo2Rc), &nBytesReturned, OTX_TASK_SYNC, OTX_TIME_INFINITY); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);
      
      DemoLogf("Done (nDemo2Rc=%d, nBytesReturned=%d).\n", nDemo2Rc, nBytesReturned);
    }
    break;
  
  case 'g':
    {
      Demo2IoInSetTxGainS sIn;
      OTX_UINT32 nBytesReturned;
      OTX_UINT16 nDemo2Rc;
      char strBuf[12];

      DemoLogf("Set Handset speaker gain for User App #%d \n", nCurrentUserApp);
      DemoLogf("Please enter gain (-3 <= gain < 7): ");
      OsaConsoleGetStr(strBuf);

      sIn.nTxGain = (OTX_INT16) strtol(strBuf, NULL, 0);

      // The gain range -3 to 7 is arbitrarily range. The DSP code will only use
      // nTxGain to perform an arithmetic shift of the 16-bit value. There is no
      // overflow checking. The range choses in this example is merely a reasonable
      // range for this particular application.
      if ((sIn.nTxGain < -3) || (sIn.nTxGain > 7)) {
        DemoLogf("Error: Gain is out of range\n");
      } else {

        DemoLogf("Setting gain to %+d\n", (OTX_INT16)sIn.nTxGain);
        result = OtxDspIoControl(g_hLogicDevs.hUserApp[nCurrentUserApp], DEMO2_IOCTL_SET_TX_GAIN, &sIn, sizeof(sIn), &nDemo2Rc, sizeof(nDemo2Rc), &nBytesReturned, OTX_TASK_SYNC, OTX_TIME_INFINITY); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);

        DemoLogf("Done (nDemo2Rc=%d, nBytesReturned=%d).\n", nDemo2Rc, nBytesReturned);
      }
    }
    break;


  case 'e':
    {
      DemoLogf("Enable Handset loop for User App #%d ...", nCurrentUserApp);
      result = OtxDrvEnable(g_hLogicDevs.hUserApp[nCurrentUserApp],   OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);
      DemoLogf("Done.\n");
    }
    break;

  case 'E':
    {
      DemoLogf("Disable Handset loop for User App #%d ...", nCurrentUserApp);
      result = OtxDrvDisable(g_hLogicDevs.hUserApp[nCurrentUserApp],   OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);
      DemoLogf("Done.\n");
    }
    break;

  case 'r':
    {
      DemoLogf("Reset Debug Struct User App #%d ...", nCurrentUserApp);
      result = OtxDrvReset(g_hLogicDevs.hUserApp[nCurrentUserApp]); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);
      DemoLogf("Done.\n");
    }
    break;

  case 'd':
    {
      OTX_UINT32 nBytesReturned;
      DemoLogf("Get Debug Data for User App #%d (non-blocking call)...", nCurrentUserApp);

      result = OtxDspIoControl(g_hLogicDevs.hUserApp[nCurrentUserApp], DEMO2_IOCTL_GET_DEBUG_DATA, NULL, 0, NULL, 0, &nBytesReturned, DEMO_TASK_REF_DEBUG, 5000); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);

      DemoLogf("Done (nBytesReturned=%d).\n", nBytesReturned);
    }
    break;

  case 's':
    {
      OTX_UINT32 nBytesReturned;
      Demo2IoOutGetSampleS dataS;

      DemoLogf("Get sample value (16-bit) from the Handset Mic for User App #%d \n", nCurrentUserApp);

      result = OtxDspIoControl(g_hLogicDevs.hUserApp[nCurrentUserApp], DEMO2_IOCTL_GET_SAMPLE, NULL, 0, &dataS, sizeof(dataS), &nBytesReturned, OTX_TASK_SYNC, OTX_TIME_INFINITY); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);

      if (OTX_SUCCESS(result)) {
        if (nBytesReturned != sizeof(Demo2IoOutGetSampleS)) {
          DemoLogf("Unexpected data length for DEMO2_IOCTL_GET_SAMPLE (nBytesReturned=%d)\n", nBytesReturned);
        } else if (dataS.nResult != OTXDSP_SDK_S_OK) {
          DemoLogf("Demo2 Error Code: %d\n", dataS.nResult);
        } else {
          DemoLogf("DataSample = %+.6d\n", (short)dataS.nDataSample);
        }
      }

      DemoLogf("Done (nBytesReturned=%d).\n", nBytesReturned);
    }
    break;

  case 'x':
    {
      DemoLogf("Disconnecting User App #%d ...", nCurrentUserApp);
      result = OtxDrvDisconnectLogicalDevice(g_hPipes.hCodecDsp[nCurrentUserApp],    -1, g_hLogicDevs.hUserApp[nCurrentUserApp],   -1); DemoCheckRc(g_hPipes.hCodecDsp[nCurrentUserApp], result);
      result = OtxDrvDisconnectLogicalDevice(g_hLogicDevs.hUserApp[nCurrentUserApp], -1, g_hPipes.hDspCodec[nCurrentUserApp], -1); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);
      DemoLogf("Done.\n");
    }
    break;

  case '0':
  case '1':
  case '2':
  case '3':
    // Select current DSP
    nCurrentUserApp = commandChar - '0';
    DemoLogf("DSP User Application #%d Selected ....\n", nCurrentUserApp);
    break;

  case 'q':
  case 'Q':
    return 1;
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}


static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *pEnv)
/*******************************************************************************
  @func
    Parse the program's command line arguments and calls upon
    requested actions.

    There are some optional switches that can be used following
    the driver name on the command line. Switches are are
    separated by spaces, and can appear in arbitrary order.
    Switches consist of a hyphen and a letter:
*******************************************************************************/
{
  short  i;

  // Set reasonable defaults before reading from command line
  pEnv->logFileName  = DEMO_DEFAULT_LOG_FILE;
  pEnv->nBoardType   = OTX_DEVICE_THOR_PCI;
  pEnv->nDsps        = 4;
  pEnv->bAsmPresent  = OTX_TRUE;
  pEnv->boardNo      = DEMO_DEFAULT_BOARD_NO;

  for (i = 1; i < argc; i++) {        // Scan command line arguments
    OsaStdStrUpr(argv[i]);
    if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
      if (*(argv[i]+1) == 'L') {   // Log file
        pEnv->logFileName = argv[i]+2;
      } else if (*(argv[i]+1) == 'N') {   // Board number
        pEnv->boardNo = strtoul(argv[i]+2, NULL, 0);
      } else if (*(argv[i]+1) == 'B') {   // Board Type
        if ( *(argv[i]+2) == 'A') {
          pEnv->nBoardType = OTX_DEVICE_ARNI_PCI;
        } else if ( *(argv[i]+2) == 'B') {
          pEnv->nBoardType = OTX_DEVICE_BALDER_S_PCI;
        } else if ( *(argv[i]+2) == 'T') {
          pEnv->nBoardType = OTX_DEVICE_THOR_PCI;
        } else if ( *(argv[i]+2) == 'V') {
          pEnv->nBoardType  = OTX_DEVICE_VIDAR_5x_PCI;
          pEnv->nDsps       = 16;
          pEnv->bAsmPresent = OTX_FALSE;
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
        exit(1);
      }
    } else {
      DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
      DemoHelpPanel();
      OsaConsoleExit();
      exit(1);
    }
  }

}


static void DemoHelpPanel()
/*******************************************************************************
  @func
    Displays a help for the command line.
*******************************************************************************/
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /L<fileName>   : Log file name\n");
  DemoLogf(" /N<boardNo>    : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /B[A|B|T|V]    : Board Type used for the demo, where\n");
  DemoLogf("                  A = Use Arni-x-PCI Board with Vidar-5x4-ASM\n");
  DemoLogf("                  B = Use Balder-8S-PCI Board with Vidar-5x4-ASM\n");
  DemoLogf("                  T = Use Thor-x-PCI Board with Vidar-5x4-ASM\n");
  DemoLogf("                  V = Use Vidar-5x16-PCI Board\n");
}



static void DemoLogf(char *formatStr, ...)
/*******************************************************************************
  @func
    Displays a message on the screen and prints it to the logfile.
*******************************************************************************/
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);
}



int main(int argc, char *argv[])
/*******************************************************************************
*******************************************************************************/
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;

  OsaConsoleInit();

  DemoLogf("DspDemo2, %s, Odin TeleSystems Inc.\n", __DATE__);
  DemoLogf("DSP SDK Demo Program for OTX PCI boards.\n");
  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
    exit(-1);
  }

  DemoLogf("Log file is %s\n", g_demoEnv.logFileName);
  DemoLogf("Board number %d\n", g_demoEnv.boardNo);

  // Connect to the Driver
  DemoLogf("Connecting lib...");
  result = OtxDrvConnectLib(); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("Creating event queue...\n");
  DemoCreateEventQueue();

  // Connect (get handles) to the physical devices on the Thor-2-PCI board
  DemoLogf("Opening physical devices...\n");
  DemoOpenPhysicalDevices();

  // Enable all physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("Enabling all opened physical devices...\n");
  DemoEnablePhysicalDevices();

  // Create and get handles to the Logical devices needed by this application
  DemoLogf("Creating logical devices...\n");
  DemoCreateLogicalDevices();

  // Enable all logical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("Enabling all opened logical devices...\n");
  DemoEnableLogicalDevices();

  // Create and get handles to the Pipe devices needed by this application
  DemoLogf("Create pipes and connect the logical devices...\n");
  DemoConnectDevices();

  // Set clock source for the board
  result = OtxBrdSetClocks(g_hPhysDevs.hBoardController0, OTX_CLOCK_MODE_LOCAL, OTX_CLOCK_SOURCE_INTERNAL); DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....

  OsaThreadStart(DemoIdleFunction);

  DemoLogf("\nMain loop running. Press 'h' for help\n");
  while (!exitFlag) {
    commandChar = getch();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

  // Clean up
  OtxDrvReleaseLib();
  OsaConsoleExit();

  return 0;
}

